#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
geoname database support routines.  Add locations to geoname DB, create geoname tables and indices.   
Provides a number of methods to lookup locations by name, feature, admin ID, etc.
"""
import logging
import os
import sys
import time
from tkinter import messagebox

from geodata import Loc, DB, GeoSearch
from geodata.GeoUtil import Query, Result, Entry


class GeoDB:
    """
    geoname database routines.  
    """

    def __init__(self, db_path, show_message: bool, exit_on_error: bool, set_speed_pragmas: bool, db_limit: int):
        """
            geoname data database init. Open database if present otherwise raise error
        # Args:
            db_path: full path to database file
            show_message: If True, show messagebox to user on error
            exit_on_error: If True, exit if significant error occurs
            set_speed_pragmas: If True, set DB pragmas for maximum performance. 
            db_limit: SQL LIMIT parameter
        # Raises:
            ValueError('Cannot open database'), ValueError('Database empty or corrupt')
        """
        self.logger = logging.getLogger(__name__)
        self.show_message = show_message
        self.exit_on_error = exit_on_error
        self.max_query_results = 50
        self.total_time = 0
        self.total_lookups = 0
        #self.select_str = 'name, country, admin1_id, admin2_id, lat, lon, feature, geoid, sdx'
        self.db_path = db_path

        # See if DB exists 
        if os.path.exists(db_path):
            db_existed = True
        else:
            db_existed = False

        self.db = DB.DB(db_filename=db_path, show_message=show_message, exit_on_error=exit_on_error)
        if self.db.err != '':
            self.logger.error(f"Error! cannot open database {db_path}.")
            raise ValueError('Cannot open database')

        # If DB was initially found 
        if db_existed:
            # Run sanity test on DB
            err = self.db.test_database('name', 'main.geodata', where='name = ? AND country = ?', args=('ba', 'fr'))

            if err:
                # DB failed sanity test
                self.logger.warning(f'DB error for {db_path}')

                if show_message:
                    if messagebox.askyesno('Error',
                                           f'Geoname database is empty or corrupt:\n\n {db_path} \n\nDo you want to delete it and rebuild?'):
                        messagebox.showinfo('', 'Deleting Geoname database')
                        self.db.conn.close()
                        os.remove(db_path)
                if exit_on_error:
                    sys.exit()
                else:
                    raise ValueError('Database empty or corrupt')

        if set_speed_pragmas:
            self.db.set_speed_pragmas()
            
        self.db_limit = db_limit
        self.db.order_string = ''
        self.db.limit_string = f'LIMIT {self.db_limit}'
        self.place_type = ''
        self.s:GeoSearch.GeoSearch = GeoSearch.GeoSearch(geodb=self)
    
    def get_db_version(self) -> int:
        """
        Get schema version of database   
        #Returns: 
            schema version of database   

        """
        # If version table does not exist, this is V1
        if self.db.table_exists('version'):
            # query version ID
            query_list = [
                Query(where="version like ?",
                      args=('%',),
                      result=Result.STRONG_MATCH)]
            select_str = '*'
            row_list = []
            self.process_query_list(result_list=row_list, select_fields=select_str, from_tbl='main.version', query_list=query_list)
            if len(row_list) > 0:
                ver = int(row_list[0][1])
                self.logger.debug(f'Database Version = {ver}')
                return ver

        # No version table, so this is V1
        self.logger.debug('No version table.  Version is 1')
        return 1

    def clear_geoname_data(self):
        """
        Delete geodata table and admin table from database
        """
        for tbl in ['geodata', 'admin']:
            # noinspection SqlWithoutWhere
            self.db.delete_table(tbl)

    def copy_georow_to_place(self, row, place: Loc, fast: bool):
        """
        Copy data from DB row into place instance   
        Country, admin1_id, admin2_id, city, lat/lon, feature, geoid are updated if available   
        #Args:   
            row: georow from geoname database   
            place: Loc instance   
            fast: Currently ignored
        #Returns:   
            None.  Place instance is updated with data from georow   
        """
        place.admin1_id = ''
        place.admin2_id = ''
        place.admin1_name = ''
        place.admin2_name = ''
        place.city = ''

        place.country_iso = str(row[Entry.ISO])
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = str(row[Entry.FEAT])
        place.geoid = str(row[Entry.ID])
        place.prefix = row[Entry.PREFIX]
        place.place_type = Loc.PlaceType.CITY

        if place.feature == 'ADM0':
            place.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            place.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city = row[Entry.NAME]
            
        self.s.update_names(place)

        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        place.city = str(place.city)
        if place.city is None:
            place.city = ''

        try:
            place.score = row[Entry.SCORE]
        except IndexError:
            pass

    def process_query_list(self, result_list, select_fields, from_tbl: str, query_list: [Query], 
                           stop_on_match=False, debug=False):
        """

        Args:
            result_list: will contain matches
            select_fields: fields to select
            from_tbl: table to select from
            query_list: .where is where clause, .args are query arguments

        Returns: Result type. Result.NO_MATCH on failure.  Result_list contains matches

        """
        # Perform each SQL query in the list
        result_type = Result.NO_MATCH
        if result_list is None:
            return

        for idx, query in enumerate(query_list):
            start = time.time()

            row_list = self.db.select(select_fields, query.where, from_tbl,
                                      query.args)
            #if debug:
                #self.logger.debug(f'{idx}) SELECT from {from_tbl} where {query.where} val={query.args} ')
                #for row in row_list:
                #    self.logger.debug(f'   FOUND {row}')
            result_list.extend(row_list)

            if len(result_list) > 0:
                result_type = query.result
                if stop_on_match:
                    break
            #else:
            #    self.logger.debug('     NO MATCH')

            elapsed = time.time() - start
            self.total_time += elapsed
            self.total_lookups += 1
            if elapsed > .1:
                self.logger.debug(f'Slow lookup. Time={elapsed:.4f}  '
                                  f'len {len(result_list)} from {from_tbl} '
                                  f'where {query.where} val={query.args} ')

            if len(result_list) > self.max_query_results:
                self.logger.debug('MAX QUERIES HIT')
                break

        return result_type

    def get_row_count(self) -> int:
        """
        Get row count of main.geodata
        :return:
        """
        return self.db.get_row_count('main.geodata')

    def close(self):
        """
        Close database.  Set optimize pragma
        """
        self.db.set_optimize_pragma()
        self.logger.info('Closing Database')
        self.db.conn.close()




""" Command argument parser that throws exception rather than exitting """

import argparse


# Variant of ArgParse which will not exit when there is a parser error, instead raises exception
class ArgumentParserError(Exception):
    pass


class ArgumentParserNoExit(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)



# -*- coding: utf-8 -*-
#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
""" Build database for geonames.org data """
import csv
import logging
import os
import sys
import time
from collections import namedtuple
from tkinter import messagebox
from typing import Dict

from geodata import GeoUtil, Loc, Country, GeoSearch, Normalize, CachedDictionary, AlternateNames, GeoDB
from geodata.GeoUtil import Entry

DB_REBUILDING = -1


class GeodataBuild:
    """
    Read in geonames.org geo files, filter them and place the entries into a sqlite db.

    The files must be downloaded from geonames.org and used according to their usage rules.
    
    geoname files: allCountries.txt,  alternateNamesV2.txt

    The geoname file is filtered to only include the countries and feature codes specified in: 1) the country_list,
      2) the feature_list, and 3) the language list.  The import also creates new Feature codes: PP1M, P1HK, and P10K using   
     geoname population data (Pop > 1M, Pop > 100K, Pop > 10K).   

    """

    def __init__(self, directory: str, display_progress,
                 show_message:bool, exit_on_error:bool, languages_list_dct:{}, feature_code_list_dct:{}, supported_countries_dct:{}):
        """
        Read in datafiles needed for geodata, filter them and create a sql db.
        Filter dictionary examples:   
            languages_list_dct={'fr','de'}
            feature_code_list_dct={'PPL', 'ADM1', 'CSTL'}
            supported_countries_dct = {'us','gb','at'}
        # Args:
            directory: base directory
            display_progress: None or Handler called with percent_done:int, msg:str
            show_message: True to show message boxes to user on errors
            exit_on_error:  True to exit on serious errors
            languages_list_dct: dictionary containing the ISO-2 languages  to load from alternateNames
            feature_code_list_dct: dictionary containing the Geonames.org feature codes to load
            supported_countries_dct: dictionary containing the ISO-2 countries to load
        """
        self.logger = logging.getLogger(__name__)
        self.geodb:[GeoDB.GeoDB, None] = None
        self.show_message = show_message
        self.geoid_main_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.geoid_admin_dict = {}  # Key is GEOID, Value is DB ID for entry
            
        self.exit_on_error = exit_on_error
        self.required_db_version = 4
        # Message to user upgrading from earlier DB version  
        self.db_upgrade_text = 'Renamed column to Feature'
        self.directory: str = directory
        self.progress_bar = display_progress
        self.line_num = 0
        self.cache_changed: bool = False
        sub_dir = GeoUtil.get_cache_directory(self.directory)
        self.country = None
        self.languages_list_dct = languages_list_dct
        self.feature_code_list_dct = feature_code_list_dct
        self.supported_countries_dct = supported_countries_dct
        self.lang_list = []

        for item in self.languages_list_dct:
            self.lang_list.append(item)

        if not os.path.exists(sub_dir):
            self.logger.warning(f'Directory] {sub_dir} NOT FOUND')
            if self.show_message:
                messagebox.showwarning('Folder not found', f'Directory\n\n {sub_dir}\n\n NOT FOUND')
            if exit_on_error:
                sys.exit()

        # Read in Text Replacement dictionary pickle - this has output text replacements
        self.output_replace_cd = CachedDictionary.CachedDictionary(sub_dir, "output_list.pkl")
        self.output_replace_cd.read()
        self.output_replace_dct: Dict[str, str] = self.output_replace_cd.dict
        self.output_replace_list = []

        for item in self.output_replace_dct:
            self.output_replace_list.append(item)

        self.entry_place = Loc.Loc()

        # Support for Geonames AlternateNames file.  Adds alternate names for entries
        self.alternate_names = AlternateNames.AlternateNames(directory=self.directory, geo_build=self,
                                                             progress_bar=self.progress_bar, filename='alternateNamesV2.txt',
                                                             lang_list=self.lang_list)

    def create_geonames_database(self)->bool:
        """
        Create geonames database from geonames.org files.
        You must call self.geodb = GeoDB.GeoDB(...) before this
        # Returns:
            True if error
        """
        # todo determine why some entries are duplicates
        # DB didnt exist.  Create tables.
        if self.geodb is None:
            self.logger.error('Cannot create DB: geodb is None')
            return True

        self.create_tables()
        
        # Set DB version to DB_REBUILDING until DB is  built, then set proper version
        self.insert_version(DB_REBUILDING)

        # Add country names
        self.country = Country.Country(progress=self.progress_bar, geo_files=self, lang_list=self.lang_list)
        self.country.add_country_names_to_db(geobuild=self)

        # Add historic names
        self.country.add_historic_names_to_db(geobuild=self)
        
        start_time = time.time()

        # Add geonames.org country files (or allCountries.txt)
        file_count = 0
        for fname in ['allCountries.txt', 'ca.txt', 'gb.txt', 'de.txt', 'fr.txt', 'nl.txt']:
            # Read  geoname files
            error = self._add_geoname_file_to_db(fname)  # Read in info (lat/long) for all places from
            if error:
                self.logger.warning(f'geoname file {fname} not found')
            else:
                file_count += 1

        if file_count == 0:
            self.logger.error(f'No geonames files found in {os.path.join(self.directory, "*.txt")}')
            return True
        
        self.logger.info(f'Geonames.org files done.  Elapsed ={(time.time() - start_time):.0f} seconds')
        
        # Create Main Indices
        self.progress("3) Step 3 of 4: Creating Indices for Database...", 95)
        start_time = time.time()
        self.create_geoid_index()
        self.create_main_indices()
        self.logger.debug(f'Indices done.  Elapsed = {(time.time() - start_time):.0f} seconds')

        # Add geonames.org alternate names
        self.progress("4) Step 4 of 4: Adding Alternate Names to Database...", 95)

        start_time = time.time()
        err = self.alternate_names.add_alternate_names_to_db()
        if err:
            self.logger.warning(f'Error reading Alternate names.')
        else:
            self.logger.info(f'Alternate names done.  Elapsed ={(time.time() - start_time):.0f} seconds')
            
        self.create_alt_indices()
        self.logger.info(f'Geonames entries = {self.geodb.get_row_count():,}')

        # Add aliases
        Normalize.add_aliases_to_db(self)

        # Done - Set Database Version
        self.insert_version(self.required_db_version)
        return err

    def _add_geoname_file_to_db(self, file) -> bool:
        """
            Read in geonames files and build lookup structure

            Read a geoname.org places file and create a db of all the places.
            1. The db contains: Name, Lat, Long, district1ID (State or Province ID),
            district2_id, feat_code

            2. Since Geonames supports over 25M entries, the db is filtered to only the countries and feature types we want
        # Args:
            file: filename (not path) for DB file.  Will be placed in directory from init call

        Returns:
            True if error
        """
        Geofile_row = namedtuple('Geofile_row',
                                 'id name name_asc alt lat lon feat_class feat_code iso iso2 admin1_id'
                                 ' admin2_id admin3_id admin4_id pop elev dem timezone mod')
        self.line_num = 0
        self.progress("Reading {}...".format(file), 0)
        path = os.path.join(self.directory, file)

        if os.path.exists(path):
            fsize = os.path.getsize(path)
            bytes_per_line = 128  # Approximate bytes per line for progress indicator
            with open(path, 'r', newline="", encoding='utf-8', errors='replace') as geofile:
                self.progress("Building Database from {}".format(file), 2)  # initialize progress bar
                reader = csv.reader(geofile, delimiter='\t')
                self.geodb.db.begin()

                # Map line from csv reader into GeonameData namedtuple
                for line in reader:
                    self.line_num += 1
                    if self.line_num % 30000 == 0:
                        # Periodically update progress
                        prog = self.line_num * bytes_per_line * 100 / fsize
                        self.progress(msg=f"1) Building Database from {file}            {prog:.1f}%", val=prog)
                    try:
                        geoname_row = Geofile_row._make(line)
                    except TypeError:
                        self.logger.error(f'Unable to parse geoname location info in {file}  line {self.line_num}')
                        continue

                    # Only handle line if it's for a country we follow and its
                    # for a Feature tag we're interested in
                    if geoname_row.iso.lower() in self.supported_countries_dct and \
                            geoname_row.feat_code in self.feature_code_list_dct:
                        self.insert_georow(geoname_row)
                        if geoname_row.name.lower() != Normalize.normalize(geoname_row.name, remove_commas=True):
                            self.insert_alternate_name(geoname_row.name,
                                                             geoname_row.id, 'ut8')

            self.progress("Write Database", 90)
            self.geodb.db.commit()
            self.progress("Database created", 100)
            return False
        else:
            return True
        
    def insert(self, geo_tuple: (), feat_code: str):
        """
        Insert a geo_row into geonames database   
        #Args:   
            geo_row: row to insert   
            feat_code: Geonames feature code of item   
        #Returns:   
            row_id for inserted row   
        """
        # We split the data into 2  tables, 1) admin: ADM0/ADM1,  and 2) geodata:  all other place types (city, county, ADM2, etc)
        if feat_code == 'ADM1' or feat_code == 'ADM0':
            sql = ''' INSERT OR IGNORE INTO admin(name,country, admin1_id,admin2_id,lat,lon,feature, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.geodb.db.execute(sql, geo_tuple)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_admin_dict[geo_tuple[Entry.ID]] = row_id
        else:
            sql = ''' INSERT OR IGNORE INTO geodata(name, country, admin1_id, admin2_id, lat, lon, feature, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.geodb.db.execute(sql, geo_tuple)
            # Add name to dictionary.  Used by AlternateNames for faster lookup during DB build
            self.geoid_main_dict[geo_tuple[Entry.ID]] = row_id

        if 'beconshire' in geo_tuple:
            self.logger.debug(f'insert [{sql}] [{geo_tuple}')
        return row_id
        
    def open_geodb(self, repair_database: bool, query_limit:int) -> bool:
        """
         Open Geoname DB file - this is the db of geoname.org city files and is stored in cache directory under geonames_data.
         The db only contains important fields and only for supported countries.
         If the db doesn't exist and repair flag is True, read the geonames.org files and build DB.   
         The DB has a version table for the schema version.  If the schema changes, the version should be updated.   
         This will check DB schema version and rebuild DB if version is out of date.   
        # Args:   
            repair_database: If True, rebuild database if error or missing   
        Returns:   
            True if error   
        """

        # Use db if it exists and has data and is correct version
        cache_dir = GeoUtil.get_cache_directory(self.directory)
        db_path = os.path.join(cache_dir, 'geodata.db')

        self.logger.debug(f'path for geodata.db: {db_path}')
        err_msg = ''

        # Validate Database setup
        if os.path.exists(db_path):
            # DB was Found
            self.logger.debug(f'DB found at {db_path}')
            self.geodb = GeoDB.GeoDB(db_path=db_path, 
                                     show_message=self.show_message, exit_on_error=self.exit_on_error,
                                     set_speed_pragmas=True, db_limit=query_limit)

            # Make sure DB is correct version
            ver = self.geodb.get_db_version()
            if ver != self.required_db_version:
                # Bad DB version 
                if ver == DB_REBUILDING:
                    # DB didn't complete rebuild
                    err_msg = f'Database only partially built.  Deleting and will rebuild on next startup'
                else:
                    # DB is out of date
                    err_msg = f'Database version will be upgraded:\n\n{self.db_upgrade_text}\n\n' \
                        f'Upgrading database from V {ver} to V {self.required_db_version}.'
                self.geodb.close()
                os.remove(db_path)
                self.logger.info(err_msg)
                if self.show_message:
                    messagebox.showinfo('Database Deleted. Will rebuild on start up', err_msg)
                sys.exit()
        else:
            err_msg = f'Database not found at\n\n{db_path}.\n\nBuilding DB'

        self.logger.debug(f'{err_msg}')
        if err_msg == '':
            # No DB errors detected
            count = self.geodb.get_row_count()
            self.logger.info(f'Geoname database has {count:,} entries\n'
                             f'------------------------------------------------------------\n')
        else:
            # DB error detected - rebuild database if flag set
            if self.show_message:
                messagebox.showinfo('Database Error', err_msg)

            self.logger.debug(err_msg)

            if repair_database:
                if os.path.exists(db_path):
                    self.geodb.close()
                    os.remove(db_path)
                    self.logger.info('Database deleted')
                    if self.show_message:
                        messagebox.showinfo('Database Deleted. Will rebuild on start up', err_msg)

                self.geodb = GeoDB.GeoDB(db_path=db_path,
                                                 show_message=self.show_message, exit_on_error=self.exit_on_error,
                                                 set_speed_pragmas=True, db_limit=query_limit)
                return self.create_geonames_database()
        return False

    @staticmethod
    def update_geo_row_name(geo_row:[], name:str, normalize=True):
        """
            Update the name entry and soundex entry with a new location name
        #Args:
            geo_row: 
            name: location name
        """
        if normalize:
            geo_row[GeoSearch.Entry.NAME] = Normalize.normalize(name, remove_commas=True)
        else:
            geo_row[GeoSearch.Entry.NAME] = name.lower()

        geo_row[GeoSearch.Entry.SDX] = GeoSearch.get_soundex(geo_row[GeoSearch.Entry.NAME])

    def insert_georow(self, geoname_row):
        """
            Create a Geo_row and insert it into DB
            This also creates new Feature codes PP1M, P1HK, and P10K using population data
        Args:
            geoname_row: ('paris', 'fr', '07', '012', 12.345, 45.123, 'PPL', '34124')

        Returns:
            None
        """
        geo_row = [None] * GeoSearch.Entry.MAX
        self.update_geo_row_name(geo_row=geo_row, name=geoname_row.name)

        geo_row[GeoSearch.Entry.ISO] = geoname_row.iso.lower()
        geo_row[GeoSearch.Entry.ADM1] = geoname_row.admin1_id
        geo_row[GeoSearch.Entry.ADM2] = geoname_row.admin2_id
        geo_row[GeoSearch.Entry.LAT] = geoname_row.lat
        geo_row[GeoSearch.Entry.LON] = geoname_row.lon
        geo_row[GeoSearch.Entry.FEAT] = geoname_row.feat_code
        geo_row[GeoSearch.Entry.ID] = geoname_row.id
            
        # Simplify feature type for Abbey/Priory, Castle, and Church.  Set feature based on population
        geo_row[GeoSearch.Entry.FEAT] = Normalize.normalize_features(feature=geo_row[GeoSearch.Entry.FEAT],
                                                                     name=geo_row[GeoSearch.Entry.NAME], pop=int(geoname_row.pop))
            
        self.insert(geo_tuple=geo_row, feat_code=geoname_row.feat_code)

        # Add abbreviations for USA states
        if geo_row[GeoSearch.Entry.ISO] == 'us' and geoname_row.feat_code == 'ADM1':
            self.update_geo_row_name(geo_row=geo_row, name=geo_row[GeoSearch.Entry.ADM1])
            self.insert(geo_tuple=geo_row, feat_code=geoname_row.feat_code)

    def get_supported_countries(self) -> [str, int]:
        """ Convert list of supported countries into sorted string """
        nm_msg = ""
        for ky in self.supported_countries_dct:
            nm_msg += ky.upper() + ', '
        return nm_msg, len(self.supported_countries_dct)

    def progress(self, msg, val):
        """ Update progress bar if there is one """
        if self.progress_bar is not None:
            self.progress_bar(val, msg)

        # If we're past 80% log item as info, otherwise log as debug
        if val > 90 or val < 10 or (val > 40 and val < 50):
            self.logger.info(f'{val:.1f}%  {msg}')
        else:
            self.logger.debug(f'{val:.1f}%  {msg}')

    @staticmethod
    def make_georow(name: str, iso: str, adm1: str, adm2: str, lat: float, lon: float, feat: str, geoid: str, sdx: str) -> ():
        """
        Create a georow based on arguments
        # Args:
            name:   
            iso:   
            adm1: admin1 id   
            adm2: admin2_id   
            lat:   
            lon: 
            feat:     
            geoid:   
            sdx:   

        # Returns:    
            georow

        """
        res = (name, iso, adm1, adm2, lat, lon, feat, geoid, sdx)
        return res

    def create_tables(self):
        """
        Create all the tables needed for the geoname database
        """
        # name, country, admin1_id, admin2_id, lat, lon, feature, geoid
        sql_geodata_table = """CREATE TABLE IF NOT EXISTS geodata    (
                id           integer primary key autoincrement not null,
                name     text COLLATE BINARY,
                country     text COLLATE BINARY,
                admin1_id     text COLLATE BINARY,
                admin2_id text COLLATE BINARY,
                lat      text  ,
                lon       text ,
                feature      text COLLATE BINARY,
                geoid      text COLLATE BINARY,
                sdx     text COLLATE BINARY
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, feature, geoid
        sql_admin_table = """CREATE TABLE IF NOT EXISTS admin    (
                id           integer primary key autoincrement not null,
                name     text COLLATE BINARY,
                country     text COLLATE BINARY,
                admin1_id     text COLLATE BINARY,
                admin2_id text COLLATE BINARY,
                lat      text  ,
                lon       text ,
                feature      text COLLATE BINARY,
                geoid      text COLLATE BINARY,
                sdx     text COLLATE BINARY
                                    );"""

        # name, lang, geoid
        sql_alt_name_table = """CREATE TABLE IF NOT EXISTS altname    (
                id           integer primary key autoincrement not null,
                name     text,
                lang     text,
                geoid      text COLLATE BINARY,
                sdx     text COLLATE BINARY
                                    );"""

        # version
        sql_version_table = """CREATE TABLE IF NOT EXISTS version    (
                id           integer primary key autoincrement not null,
                version     integer
                                    );"""

        for tbl in [sql_geodata_table, sql_admin_table, sql_version_table, sql_alt_name_table]:
            self.geodb.db.create_table(tbl)

    def insert_alternate_name(self, alternate_name: str, geoid: str, lang: str):
        """
        Add alternate name to altname table
        #Args:   
            alternate_name: alternate name to add for this geoid   
            geoid: geonames.org geoid   
            lang: ISO lang code for this entry   

        #Returns: None   

        """
        sdx = GeoSearch.get_soundex(Normalize.normalize(alternate_name, True))
        row = (alternate_name, lang, geoid, sdx)
        sql = ''' INSERT OR IGNORE INTO altname(name,lang, geoid, sdx)
                  VALUES(?,?,?,?) '''
        self.geodb.db.execute(sql, row)

    def insert_version(self, db_version: int):
        """
        Insert DB version into Database.  This is used to track when DB schema changes   
        #Args:   
            db_version: Version of this DB schema   

        #Returns: None   

        """
        self.geodb.db.begin()

        # Delete previous version from table
        # noinspection SqlWithoutWhere
        sql = '''DELETE FROM version;'''
        args = None
        self.geodb.db.execute(sql, args)

        sql = ''' INSERT OR IGNORE INTO version(version)
                  VALUES(?) '''
        args = (db_version,)
        self.geodb.db.execute(sql, args)
        self.geodb.db.commit()

    def create_geoid_index(self):
        """
        Create database indices for GEOID
        """
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS geoid_idx ON geodata(geoid  )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS admgeoid_idx ON admin(geoid  )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS altnamegeoid_idx ON altname(geoid  )')

    def create_main_indices(self):
        """
        Create indices for geoname database
        """
        # Indices for geodata table
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS name_idx ON geodata(name  , country, admin1_id  )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS admin1_idx ON geodata(admin1_id   )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS sdx_idx ON geodata(sdx  , country   )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_admin2_idx ON geodata(admin1_id  , feature  , '
                                                    'admin2_id  )')

        # Indices for admin table
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_name_idx ON admin(name  , country  )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_admin1_idx ON admin(admin1_id  , feature  )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_country_idx ON admin(country  , feature  )')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_sdx_idx ON admin(sdx  )')

    def create_alt_indices(self):
        # Indices for altname table
        self.logger.debug('create alt index')
        self.geodb.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS alt_sdx_idx ON altname(sdx  )')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
Provides a place lookup gazeteer based on files from geonames.org.   
  Provides the location lookup methods for the geodata package.  
           
+ Creates a local sqlite3 place database of geonames.org data  
+ Parses lookup text and returns multiple matches ranked by closeness to lookup term  
+ Provides latitude/longitude  
+ Supports Wildcard search, Phonetic/Soundex search, and Word search of place database names  
+ Search by feature type (e.g. mountain, cemetery, palace, etc)  
+ Database can be filtered to only include specified countries, languages, and feature types   
   
   Main routines for Geodata package:   
    example.py - a sample demonstrating place lookups using geodata   
    geodata.open - open database.  create DB if missing   
    geodata.find_best_match - parse location and provide the best match   
    geodata.find_matches - parse location and provide a ranked list of matches   
    geodata.find_feature - lookup location by feature type and provide a ranked list of matches   
    normalize.py - Normalize text for lookup
  """
import collections
import copy
import logging
from operator import itemgetter

from geodata import GeoUtil, GeodataBuild, Loc, MatchScore, GeoSearch


class Geodata:
    """
Provide a place lookup gazeteer based on files from geonames.org  
   
    """

    def __init__(self, directory_name: str, display_progress,
                 show_message: bool, exit_on_error: bool, languages_list_dct, feature_code_list_dct,
                 supported_countries_dct):
        """
            Init

        #Args:
            directory_name: directory where geoname.org files are.  DB will be in 'cache' folder under this   
            display_progress: None or function to display progress(percent_done:int, msg:str)  
            show_message: If True - show TKInter message dialog on error   
            exit_on_error: If True - exit on significant error   
            languages_list_dct: Dictionary of ISO-2 languages to import from AlternateNamesV2.txt   
            feature_code_list_dct: Dictionary of Geoname Feature codes to import into DB   
            supported_countries_dct: Dictionary of ISO-2 Country codes to import into DB   
        """
        self.logger = logging.getLogger(__name__)
        self.display_progress = display_progress
        self.save_place: Loc = Loc.Loc()
        self.miss_diag_file = None
        self.distance_cutoff = 0.6  # Value to determine if two lat/longs are similar based on Rectilinear Distance
        self.geo_build = GeodataBuild.GeodataBuild(str(directory_name), display_progress=self.display_progress,
                                                   show_message=show_message, exit_on_error=exit_on_error,
                                                   languages_list_dct=languages_list_dct,
                                                   feature_code_list_dct=feature_code_list_dct,
                                                   supported_countries_dct=supported_countries_dct)

    def find_matches(self, location: str, place: Loc, plain_search) -> GeoUtil.Result:
        """
            Find a location in the geoname database.  On successful match, place.georow_list will contain   
            a list of georows that matched the name.  Each georow can be copied to a Loc structure by   
            calling process_result   

        #Args:   
            location: comma separated name of location to find, e.g. 'Los Angeles, California, USA'   
            place: Loc structure   
            plain_search: If True then don't do wildcard searches   
        #Returns:   
            GeoUtil.Result code   
        """
        place.parse_place(place_name=location, geo_db=self.geo_build.geodb)

        self.is_country_valid(place)
        if place.result_type == GeoUtil.Result.NOT_SUPPORTED:
            return place.result_type

        # Create full entry text
        place.update_names(self.geo_build.output_replace_dct)
        
        flags = ResultFlags(limited=False, filtered=False)
        result_list = []  # We will do different search types and append all results into result_list

        # self.logger.debug(f'== FIND LOCATION City=[{place.city}] Adm2=[{place.admin2_name}]\
        # Adm1=[{place.admin1_name}] Pref=[{place.prefix}] Cntry=[{place.country_name}] iso=[{place.country_iso}]  Type={place.place_type} ')

        # Save a shallow copy of place so we can restore fields
        self.save_place = copy.copy(place)

        # After parsing, last token is either country or underscore. 
        # Second to last is either Admin1 or underscore
        # If >2 tokens:  token[0] is placed in City and in Prefix
        # If >3 tokens:  token[1] is placed in Admin2 and appended to Prefix

        # 1) Try lookup based on standard parsing: lookup city, county, state/province, or country as parsed
        self.logger.debug(f'  1) Standard, based on parsing.  pref [{place.prefix}] city [{place.city}]'
                          f' sdx={GeoSearch.get_soundex(place.city)} '
                          f'feat={place.feature}')
        
        if place.place_type != Loc.PlaceType.COUNTRY and place.place_type != Loc.PlaceType.ADMIN2:
            self.geo_build.geodb.s.lookup_place(place=place)
            #self.logger.debug(place.georow_list)
            if place.georow_list:
                result_list.extend(place.georow_list)
            #self.logger.debug(result_list)

            # Restore fields
            self._restore_fields(place, self.save_place)
    
            # 2) Try second token (Admin2) as a city 
            if place.admin2_name != '': # and len(result_list) < 9:
                place.georow_list.clear()
                self._find_type_as_city(place, Loc.PlaceType.ADMIN2)
                #self.logger.debug(place.georow_list)
    
                if place.georow_list:
                    result_list.extend(place.georow_list)
                    #self.logger.debug(result_list)
                self._restore_fields(place, self.save_place)
    
            #  Move result_list into place georow list
            place.georow_list.clear()
            place.georow_list.extend(result_list)
            #self.logger.debug(place.georow_list)
        else:
            #self.logger.debug('ignore country, admin2')
            pass

        if len(place.georow_list) > 0:
            #self.logger.debug('process results')
            self.process_results(place=place, flags=flags)
            flags = self.filter_results(place)
        #self.logger.debug(place.georow_list)

        if len(place.georow_list) == 0:
            # NO MATCH
            self.logger.debug(f'Not found.')
            if place.result_type != GeoUtil.Result.NO_COUNTRY and place.result_type != GeoUtil.Result.NOT_SUPPORTED:
                place.result_type = GeoUtil.Result.NO_MATCH
        elif len(place.georow_list) > 1:
            self.logger.debug(f'Success!  {len(place.georow_list)} matches')
            place.result_type = GeoUtil.Result.MULTIPLE_MATCHES

        # Process the results
        self.process_results(place=place, flags=flags)
        # self.logger.debug(f'Status={place.status}')
        return place.result_type

    def find_best_match(self, location: str, place: Loc) -> bool:
        """
            Find the best scoring match for this location in the geoname dictionary.  
        #Args:  
            location:  location name, e.g. Los Angeles, California, USA   
            place:  Loc instance   
        #Returns: True if a match was found     
            place is updated with -- lat, lon, district, city, country_iso, result code  
        """
        #  First parse the location into <prefix>, city, <district2>, district1, country.
        #  Then look it up in the place db

        self.find_matches(location, place, plain_search=False)

        # Clear to just best entry
        flags = self.filter_results(place)
        # If multiple matches, truncate to first match
        if len(place.georow_list) > 0:
            place.georow_list = place.georow_list[:1]
            self.process_results(place=place, flags=flags)
            place.set_place_type()

            nm = f'{place.get_long_name(self.geo_build.output_replace_dct)}'
            print(f'Found pre=[{place.prefix}{place.prefix_commas}] Nam=[{nm}]')
            return True
        else:
            return False

    def find_geoid(self, geoid: str, place: Loc):
        """
        Lookup by geoid   
        #Args:   
            geoid:  Geonames.org geoid
            place:  Location fields in place are updated

        #Returns: None. Location fields in place are updated

        """
        flags = ResultFlags(limited=False, filtered=False)
        place.geoid = geoid
        place.georow_list.clear()
        self.geo_build.geodb.s.lookup_geoid(georow_list=place.georow_list, geoid=place.geoid, place=place)
        if len(place.georow_list) == 0:
            self.geo_build.geodb.s.lookup_geoid(georow_list=place.georow_list, geoid=place.geoid, place=place, admin=True)

        if len(place.georow_list) > 0:
            place.result_type = GeoUtil.Result.STRONG_MATCH
            self.process_results(place=place, flags=flags)
            #self.logger.debug(f'found geoid {place.georow_list[0]}')
        else:
            place.result_type = GeoUtil.Result.NO_MATCH
            #self.logger.debug(f'NOT FOUND geoid {geoid}')

    def _find_type_as_city(self, place: Loc, typ):
        """
            Do a lookup using the field specifed by typ as a city name.  E.g. if typ is PlaceType.ADMIN1 then   
            use the place.admin1_name field to do the city lookup   
        #Args:   
            place: Loc instance   
            typ: Loc.PlaceType - Specifies which field to use as target for lookup   

        #Returns:  None   
            place.georow_list is updated with matches   
        """
        # place.standard_parse = False
        typ_name = ''
        if typ == Loc.PlaceType.CITY:
            # Try City as city (do as-is)
            typ_name = 'City'
            pass
        elif typ == Loc.PlaceType.ADMIN2:
            # Try ADMIN2 as city
            if place.admin2_name != '':
                # if '*' not in place.city:
                #    place.prefix += ' ' + place.city
                place.city = place.admin2_name
                place.admin2_name = ''
                typ_name = 'Admin2'
        elif typ == Loc.PlaceType.PREFIX:
            # Try Prefix as City
            if place.prefix != '':
                place.city = place.prefix
                # if '*' not in tmp:
                #    place.prefix = tmp
                typ_name = 'Prefix'
        elif typ == Loc.PlaceType.ADVANCED_SEARCH:
            # Advanced Search
            self.geo_build.geodb.lookup_place(place=place)
            return
        else:
            self.logger.warning(f'Unknown TYPE {typ}')

        if typ_name != '':
            self.logger.debug(f'2) Try {typ_name} as City.  Target={place.city}  pref [{place.prefix}] ')

            place.place_type = Loc.PlaceType.CITY
            self.geo_build.geodb.s.lookup_place(place=place)


    def _lookup_city_as_admin2(self, place: Loc, result_list):
        """
        Lookup place.city as admin2 name   
        #Args:   
            place:     
            result_list:   

        #Returns:   

        """
        # Try City as ADMIN2
        # place.standard_parse = False
        place.admin2_name = place.city
        place.city = ''
        place.place_type = Loc.PlaceType.ADMIN2
        self.logger.debug(f'  Try admin2  [{place.admin2_name}] as city [{place.get_five_part_title()}]')
        self.geo_build.geodb.lookup_place(place=place)
        result_list.extend(place.georow_list)

    def find_feature(self, place):
        """
        Lookup location with - name, country, and feature  

        #Args:   
            place: place.name, place.country, and place.feature are used for lookup  
        #Returns:   
            None.  place.georow_list contains matches   

        """
        self.logger.debug('Feature Search')
        self._find_type_as_city(place, place.place_type)

        #if len(place.georow_list) > 0:
            # Build list - sort and remove duplicates
            # self.logger.debug(f'Match {place.georow_list}')
            #flags = ResultFlags(limited=False, filtered=False)
            #self.process_results(place=place, flags=flags)
            #self.filter_results(place)

    def process_results(self, place: Loc, flags) -> None:
        """
            Update fields in place record using first entry in place.georow_list   
            Updates fields with available data: city, admin1, admin2, country, lat/long, feature, etc.   
        #Args:    
            place: Loc instance   
            flags: Flags tuple as returned by sort_results   

        #Returns:    
            None.  place instance fields are updated   
        """
        # self.logger.debug(f'**PROCESS RESULT:  Res={place.result_type}   Georow_list={place.georow_list}')
        if place.result_type == GeoUtil.Result.NOT_SUPPORTED:
            place.place_type = Loc.PlaceType.COUNTRY

        if place.result_type in GeoUtil.successful_match and len(place.georow_list) > 0:
            self.geo_build.geodb.copy_georow_to_place(row=place.georow_list[0], place=place, fast=False)
        elif len(place.georow_list) > 0 and place.result_type != GeoUtil.Result.NOT_SUPPORTED:
            # self.logger.debug(f'***RESULT={place.result_type} Setting to Partial')
            place.result_type = GeoUtil.Result.PARTIAL_MATCH

        place.set_place_type_text()

    @staticmethod
    def distance(lat_a: float, lon_a: float, lat_b: float, lon_b: float):
        """
        Returns rectilinear distance in degrees between two lat/longs   
        Args:   
            lat_a: latitude of point A   
            lon_a: longitude of point A   
            lat_b: latitude of point B   
            lon_b: longitude of point B   
        Returns: Rectilinear distance between two points   

        """
        return abs(lat_a - lat_b) + abs(lon_a - lon_b)

    def remove_duplicates(self, place):
        # sort list by LON/LAT and score so we can remove dups
        #for row in place.georow_list:
        #    self.logger.debug(row)
        if len(place.georow_list) == 0:
            self.logger.debug('empty')
            return
        rows_sorted_by_latlon = sorted(place.georow_list, key=itemgetter(GeoUtil.Entry.LON, GeoUtil.Entry.LAT, GeoUtil.Entry.SCORE))
        place.georow_list.clear()

        # Create a dummy 'previous' row so the comparison to previous entry works on the first item
        prev_geo_row = self.geo_build.make_georow(name='q', iso='q', adm1='q', adm2='q', lat=900, lon=900, feat='q', geoid='q', sdx='q')
        georow_idx = 0

        # Keep track of list by GEOID to ensure no duplicates in GEOID
        geoid_dict = {}  # Key is GEOID.  Value is List index

        # Find and remove if two entries are duplicates - defined as two items with:
        #  1) same GEOID or 2) same name and lat/lon is within Box Distance of 0.6 degrees
        for geo_row in rows_sorted_by_latlon:
            # self.logger.debug(f'{geo_row[GeoUtil.Entry.NAME]},{geo_row[GeoUtil.Entry.FEAT]} '
            #                  f'{geo_row[GeoUtil.Entry.SCORE]:.1f} {geo_row[GeoUtil.Entry.ADM2]}, '
            #                  f'{geo_row[GeoUtil.Entry.ADM1]} {geo_row[GeoUtil.Entry.ISO]}')
            if self._valid_year_for_location(place.event_year, geo_row[GeoUtil.Entry.ISO], geo_row[GeoUtil.Entry.ADM1], 60) is False:
                # Skip location if location name  didnt exist at the time of event WITH 60 years padding
                continue

            if self._valid_year_for_location(place.event_year, geo_row[GeoUtil.Entry.ISO], geo_row[GeoUtil.Entry.ADM1], 0) is False:
                # Flag if location name  didnt exist at the time of event
                date_filtered = True

            old_row = list(geo_row)
            geo_row = tuple(old_row)

            if geo_row[GeoUtil.Entry.NAME] != prev_geo_row[GeoUtil.Entry.NAME]:
                # Add this item to georow list since it has a different name.  Also add its idx to geoid dict
                place.georow_list.append(geo_row)
                geoid_dict[geo_row[GeoUtil.Entry.ID]] = georow_idx
                georow_idx += 1
            elif geoid_dict.get(geo_row[GeoUtil.Entry.ID]):
                # We already have an entry for this geoid.  Replace it if this one has better score
                row_idx = geoid_dict.get(geo_row[GeoUtil.Entry.ID])
                old_row = place.georow_list[row_idx]
                if geo_row[GeoUtil.Entry.SCORE] < old_row[GeoUtil.Entry.SCORE]:
                    # Same GEOID but this has better score so replace other entry.  
                    place.georow_list[row_idx] = geo_row
                    self.logger.debug(f'Better score {geo_row[GeoUtil.Entry.SCORE]} < '
                                      f'{old_row[GeoUtil.Entry.SCORE]} {geo_row[GeoUtil.Entry.NAME]}')
            elif self.distance(float(prev_geo_row[GeoUtil.Entry.LAT]), float(prev_geo_row[GeoUtil.Entry.LON]),
                               float(geo_row[GeoUtil.Entry.LAT]), float(geo_row[GeoUtil.Entry.LON])) > self.distance_cutoff:
                # Add this item to georow list since Lat/lon is different from previous item.  Also add its idx to geoid dict 
                place.georow_list.append(geo_row)
                geoid_dict[geo_row[GeoUtil.Entry.ID]] = georow_idx
                georow_idx += 1
            elif geo_row[GeoUtil.Entry.SCORE] < prev_geo_row[GeoUtil.Entry.SCORE]:
                # Same Lat/lon but this has better score so replace previous entry.  
                place.georow_list[georow_idx - 1] = geo_row
                geoid_dict[geo_row[GeoUtil.Entry.ID]] = georow_idx - 1
                # self.logger.debug(f'Use. {geo_row[GeoUtil.Entry.SCORE]}  < {prev_geo_row[GeoUtil.Entry.SCORE]} {geo_row[GeoUtil.Entry.NAME]}')

            prev_geo_row = geo_row

    def filter_results(self, place: Loc):
        """
            Sort place.georow_list by match score and eliminate duplicates   
        
        In case of duplicate, keep the one with best match score.   
        See MatchScore.match_score() for details on score calculation    
        Discard names that didnt exist at time of event (update result flag if this occurs)  
        Duplicates are defined as two items with:  
        1) same GEOID or 2) same name and similar lat/lon (within Rectilinear Distance of distance_cutoff degrees)  
        
        Add flag if we hit the lookup limit  
        #Args:   
            place:   
        
        #Returns:   
            ResultFlags(limited=limited_flag, filtered=date_filtered)   
        """

        date_filtered = False  # Flag to indicate whether we dropped locations due to event date
        # event_year = place.event_year

        if len(place.georow_list) > 100:
            limited_flag = True
        else:
            limited_flag = False

        if len(place.georow_list) == 0:
            self.logger.debug('EMPTY')
            return ResultFlags(limited=limited_flag, filtered=date_filtered)

        # Remove duplicate locations in list (have same name and lat/lon)
        self.remove_duplicates(place)

        gap_threshold = 0
        score = 0

        # Sort places in match_score order
        new_list = sorted(place.georow_list, key=itemgetter(GeoUtil.Entry.SCORE, GeoUtil.Entry.ADM1))
        #self.logger.debug(new_list[0])
        min_score = new_list[0][GeoUtil.Entry.SCORE]
        place.georow_list.clear()

        # Go through sorted list and only add items to georow_list that are close to the best score
        for rw, geo_row in enumerate(new_list):
            score = geo_row[GeoUtil.Entry.SCORE]
            # admin1_name = self.geo_build.geodb.get_admin1_name_direct(geo_row[GeoUtil.Entry.ADM1], geo_row[GeoUtil.Entry.ISO])
            # admin2_name = self.geo_build.geodb.get_admin2_name_direct(geo_row[GeoUtil.Entry.ADM1],
            #                                                          geo_row[GeoUtil.Entry.ADM2], geo_row[GeoUtil.Entry.ISO])

            base = MatchScore.Score.VERY_GOOD + (MatchScore.Score.GOOD / 3)
            gap_threshold = base + abs(min_score) * .6

            # Range to display when there is a strong match
            #if (min_score <= base and score > min_score + gap_threshold) or score > MatchScore.Score.VERY_POOR * 1.5:
            if score > min_score + gap_threshold:

                self.logger.debug(f'SKIP Score={score:.1f} Min={min_score:.1f} Gap={gap_threshold:.1f} [{geo_row[GeoUtil.Entry.PREFIX]}]'
                                  f' {geo_row[GeoUtil.Entry.NAME]},'
                                  f' {geo_row[GeoUtil.Entry.ADM2]},'
                                  f' {geo_row[GeoUtil.Entry.ADM1]} ')
            else:
                place.georow_list.append(geo_row)
                self.logger.debug(f'Score {score:.1f} [{geo_row[GeoUtil.Entry.PREFIX]}] {geo_row[GeoUtil.Entry.NAME]}, '
                                  f'AD2={geo_row[GeoUtil.Entry.ADM2]},'
                                  f' AD1={geo_row[GeoUtil.Entry.ADM1]} {geo_row[GeoUtil.Entry.ISO]}')

        #self.logger.debug(f'min={min_score:.1f}, gap2={gap_threshold:.1f} strong cutoff={min_score + gap_threshold:.1f}')

        if min_score <= MatchScore.Score.VERY_GOOD and len(place.georow_list) == 1 and place.result_type != GeoUtil.Result.NOT_SUPPORTED:
            place.result_type = GeoUtil.Result.STRONG_MATCH
        else:
            # Log item that we couldnt match
            if self.miss_diag_file:
                self.miss_diag_file.write(
                    f'Lookup {place.original_entry} thresh={gap_threshold} gap={score - min_score}\n\n')

        return ResultFlags(limited=limited_flag, filtered=date_filtered)

    def open(self, repair_database: bool, query_limit: int):
        """
        Open geodb.  Create DB if needed   
        #Args:  
            repair_database: If True, create DB if missing or damaged. 
            query_limit:  SQL query limit 
        #Returns:  
            True if error  
        """
        self._progress("Reading Geoname files...", 70)
        return self.geo_build.open_geodb(repair_database=repair_database, query_limit=query_limit)

    def _progress(self, msg: str, percent: int):
        if self.display_progress is not None:
            self.display_progress(percent, msg)
        else:
            self.logger.debug(msg)

    def is_country_valid(self, place: Loc) -> bool:
        """
        See if COUNTRY is present and is in the supported country list   

        #Args:   
            place:  

        #Returns:   
            True if country is valid   
        """
        if place.country_iso == '':
            place.result_type = GeoUtil.Result.NO_COUNTRY
            is_valid = False
        elif place.country_iso not in self.geo_build.supported_countries_dct:
            self.logger.debug(f'[{place.country_iso}] not supported')
            place.result_type = GeoUtil.Result.NOT_SUPPORTED
            place.place_type = Loc.PlaceType.COUNTRY
            is_valid = False
        else:
            is_valid = True

        return is_valid

    @staticmethod
    def _valid_year_for_location(event_year: int, country_iso: str, admin1: str, pad_years: int) -> bool:
        """
        See if this state/province had modern names at the time of the event. Only US and Canada currently supported.   
        For example, looking up New York for year 1410 would be invalid since it did not have an English name at that time.   
        Data is based on https://en.wikipedia.org/wiki/List_of_North_American_settlements_by_year_of_foundation   
        Geonames has support for date ranges on names but that data is sparsely populated and not used here yet.   

        #Args:   
            event_year: Year to check   
            country_iso: ISO-2 country code   
            admin1: State/Province name   
            pad_years: Number of years to pad for inaccuracy   

        #Returns: 
            True if valid   

        """
        if not event_year:
            return True

        # Try looking up start year by state/province
        place_year = admin1_name_start_year.get(f'{country_iso}.{admin1.lower()}')
        if place_year is None:
            # Try looking up start year by country
            place_year = country_name_start_year.get(country_iso)
        if place_year is None:
            place_year = -1

        if event_year + pad_years < place_year:
            # self.logger.debug(f'Invalid year:  incorporation={place_year}  event={event_year} loc={admin1},{iso} pad={padding}')
            return False
        else:
            return True

    @staticmethod
    def _feature_priority(feature: str):
        """
        Returns 0-100 for feature priority.  PP1M - city with 1 million people is zero 

        #Args:   
            feature:   

        #Returns:   
            0-100 for feature priority   

        """
        res = feature_priority.get(feature)
        if res:
            return 100.0 - res
        else:
            return 100.0 - feature_priority.get('DEFAULT')

    def open_diag_file(self, miss_diag_fname: str):
        """
        Open diagnostic file   

        #Args:
            miss_diag_fname:  

        #Returns:   

        """
        self.miss_diag_file = open(miss_diag_fname, 'wt')

    def close_diag_file(self):
        """
        Close diagnostic file   

        Returns:   
        """
        if self.miss_diag_file:
            self.miss_diag_file.close()

    @staticmethod
    def _restore_fields(place, save_place):
        # Restore fields that were overwritten
        place.city = save_place.city
        place.admin2_name = save_place.admin2_name
        place.prefix = save_place.prefix

    def close(self):
        """
        Close files and database   

        Returns: None   

        """
        if self.geo_build:
            self.geo_build.geodb.close()

    def log_results(self, geo_row_list):
        for geo_row in geo_row_list:
            self.logger.debug(f'    {geo_row[GeoUtil.Entry.NAME]}')


# Entries are only loaded from geonames.org files if their feature is in this list
# Highest value is for large city or capital
# Also, If there are 2 identical entries, we only add the one with higher feature priority.  
# These scores are also used for match ranking score
# Note: PP1M, P1HK, P10K do not exist in Geonames and are created by geodata.geodataBuild
feature_priority = {
    'PP1M': 100, 'ADM1': 96, 'PPLA': 96, 'PPLC': 96, 'PPLH': 96, 'ADM0': 93, 'PPLA2': 93, 'P1HK': 93,
    'P10K': 89, 'PPLX': 82, 'PP1K': 82, 'PRN': 71, 'PRSH': 71, 'RLG': 71, 'RUIN': 71, 'STG': 71,
    'PPLG': 75, 'RGN': 71, 'AREA': 71, 'NVB': 71, 'PPLA3': 71, 'ADMF': 71, 'PPLA4': 69, 'PPLF': 69, 'ADMX': 66,
    'PPLQ': 60, 'PPLR': 60, 'PPLS': 55, 'PPLL': 55, 'PPLW': 55, 'PPL': 55, 'SQR': 50, 'ISL': 50,
    'ADM2': 45, 'CH': 44, 'MSQE': 44, 'MSTY': 44, 'SYG': 44, 'MUS': 44, 'CMTY': 44, 'CSTL': 44, 'EST': 44,
    'MILB': 44, 'MNMT': 44, 'PAL': 44, 'HSTS': 42, 'PRK': 42, 'ADM3': 32,
    'BTL' : 22, 'HSP': 0, 'VAL': 0, 'MT': 0, 'ADM4': 0, 'DEFAULT': 0,
    }

ResultFlags = collections.namedtuple('ResultFlags', 'limited filtered')

# Starting year this country name was valid
country_name_start_year = {
    'cu': -1,
    }

# Starting year when modern names were valid for this state/province 
# https://en.wikipedia.org/wiki/List_of_North_American_settlements_by_year_of_foundation
admin1_name_start_year = {
    'us.al': 1711,
    'us.ak': 1774,
    'us.az': 1775,
    'us.ar': 1686,
    'us.ca': 1769,
    'us.co': 1871,
    'us.ct': 1633,
    'us.de': 1638,
    'us.dc': 1650,
    'us.fl': 1565,
    'us.ga': 1566,
    'us.hi': -1,
    'us.id': 1862,
    'us.il': 1703,
    'us.in': 1715,
    'us.ia': 1785,
    'us.ks': 1870,
    'us.ky': 1775,
    'us.la': 1699,
    'us.me': 1604,
    'us.md': 1633,
    'us.ma': 1620,
    'us.mi': 1784,
    'us.mn': 1820,
    'us.ms': 1699,
    'us.mo': 1765,
    'us.mt': 1877,
    'us.ne': 1854,
    'us.nv': 1905,
    'us.nh': 1638,
    'us.nj': 1624,
    'us.nm': 1598,
    'us.ny': 1614,
    'us.nc': 1653,
    'us.nd': 1871,
    'us.oh': 1785,
    'us.ok': 1889,
    'us.or': 1811,
    'us.pa': 1682,
    'us.ri': 1636,
    'us.sc': 1663,
    'us.sd': 1865,
    'us.tn': 1739,
    'us.tx': 1685,
    'us.ut': 1847,
    'us.vt': 1650,
    'us.va': 1607,
    'us.wa': 1825,
    'us.wv': 1788,
    'us.wi': 1685,
    'us.wy': 1867,
    'ca.01': 1795,
    'ca.02': 1789,
    'ca.03': 1733,
    'ca.04': 1766,
    'ca.05': 1583,
    'ca.07': 1604,
    'ca.08': 1673,
    'ca.09': 1764,
    'ca.10': 1541,
    'ca.11': 1862,
    'ca.12': 1700,
    'ca.13': 1700,
    'ca.14': 1700
    }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""Provide SQLite database helper functions - insert, select, create tables and indices, etc"""
import logging
import sqlite3
import sys
import traceback
from tkinter import messagebox


class DB:
    """
    Sqlite3  helper functions
    """

    def __init__(self, db_filename: str, show_message: bool, exit_on_error: bool):
        """
            Initialize and create a database connection to db_filename
        # Args:
            db_filename: Database filename
            show_message: If true, show messagebox to user on errors
            exit_on_error: If true, sys exit on significant errors
        # Raises:
            ValueError('Cannot open database')
        """
        self.logger = logging.getLogger(__name__)
        self._order_str = ''
        self._limit_str = ''
        self.cur = None
        self.total_time = 0
        self.total_lookups = 0
        self.show_message = show_message
        self.exit_on_error = exit_on_error
        self.err = ''

        # create database connection
        self.conn = self._connect(db_filename=db_filename)
        if self.conn is None:
            self.err = f"Error! cannot open database {db_filename}."
            self.logger.error(f"Error! cannot open database {db_filename}.")
            raise ValueError('Cannot open database')

    def _connect(self, db_filename: str):
        """
            Create a database connection to the SQLite database

        # Args:
            db_filename: database filename
        # Returns:
            Connection object or None. Self.err is set to Exception text. Shows Messagebox and/or exits on error if flags set.
        """
        self.err = ''
        try:
            conn = sqlite3.connect(db_filename)
            self.logger.info(f'Using database {db_filename}')
            return conn
        except Exception as e:
            if self.show_message:
                messagebox.showwarning('Error', f'Database Connection Error\n {e}')
            self.err = e
            self.logger.error(e)
            if self.exit_on_error:
                sys.exit()
            else:
                return None

    @property
    def order_string(self):
        """
        Get value of  ORDER BY   
        Returns: value of  ORDER BY
        """
        return self._order_str

    @order_string.setter
    def order_string(self, order_str: str):
        """
        Set value for ORDER BY parameter

        # Args:
            order_str:   

        # Returns: None
        """
        self._order_str = order_str

    @property
    def limit_string(self):
        """
        Get value of  Limit parameter   
        Returns: value of LIMIT parameter
        """
        return self._limit_str

    @limit_string.setter
    def limit_string(self, limit_str: str):
        """
        Set value for LIMIT parameter

        # Args:
            limit_str:   

        # Returns: None
        """
        self._limit_str = limit_str

    def set_pragma(self, pragma: str):
        """
        Set a sqlite3 PRAGMA e.g. 'temp_store = memory'   
        # Args:
            pragma: pragma statement

        Returns: None
        """
        cur = self.conn.cursor()
        cur.execute(pragma)
        self.conn.commit()

    def create_table(self, create_table_sql: str):
        """
        Execute a SQL create table statement   
        # Args:
            create_table_sql: a full CREATE TABLE SQL statement   

        # Returns: True if error.  Self.err is set to Exception text. Shows Messagebox and/or exits on error if flags set.   
        # Raises: Nothing.  DB exceptions are suppressed.  
        """
        self.err = ''
        try:
            c = self.conn.cursor()
            c.execute(create_table_sql)
            self.conn.commit()
            # self.logger.debug(f'Create DB table \n{create_table_sql}')  # Print  table name for logging
        except Exception as e:
            if self.show_message:
                messagebox.showwarning('Error', e)
            self.err = e
            self.logger.error(e)
            traceback.print_stack()

            if self.exit_on_error:
                sys.exit()
            return True
        return False

    def create_index(self, create_index_sql: str):
        """
        Execute a SQL create index statement
        # Args:
            create_index_sql: a full CREATE INDEX SQL statement

        # Returns: True if error.  Self.err is set to Exception text. Shows Messagebox and/or exits on error if flags set.   
        # Raises: Nothing.  DB exceptions are suppressed. 
        """
        self.err = ''
        try:
            c = self.conn.cursor()
            c.execute(create_index_sql)
            self.conn.commit()
        except Exception as e:
            if self.show_message:
                messagebox.showwarning('Error', e)
            self.err = e
            self.logger.error(e)
            traceback.print_stack()

            if self.exit_on_error:
                sys.exit()
            return True
        return False

    def delete_table(self, tbl):
        """
        Delete table

        # Args:
            tbl:  table name
                # Returns: True if error.  Self.err is set to Exception text. Shows Messagebox and/or exits on error if flags set.   
        # Raises: Nothing.  DB exceptions are suppressed.   
        """
        self.err = ''
        cur = self.conn.cursor()
        try:
            # noinspection SqlWithoutWhere
            cur.execute(f'DELETE FROM {tbl}')
        except Exception as e:
            if self.show_message:
                messagebox.showwarning('Error', f'Database delete table error\n {e}')
            self.err = e
            self.logger.error(e)
            traceback.print_stack()

            if self.exit_on_error:
                sys.exit()
            return True
        return False

    def get_row_count(self, table_name):
        """
        Get row count of specified table
        # Args:
            table_name: 
        Returns: row count of specified table
        """
        cur = self.conn.cursor()
        cur.execute(f'SELECT COUNT(*) FROM {table_name}')
        res = cur.fetchall()
        count = res[0][0]
        return count

    def begin(self):
        """ Begin transaction """
        self.cur = self.conn.cursor()
        self.cur.execute('BEGIN')

    def execute(self, sql, args) -> int:
        """
        Execute a SQL statement
        # Args:
            sql: a full SQL statement

        # Returns: row.  Self.err is set to Exception text. Shows Messagebox and/or exits on error if flags set.   
        # Raises: Nothing.  DB exceptions are suppressed. 
        """
        self.err = ''
        try:
        #if True:
            if args:
                self.cur.execute(sql, args)
            else:
                self.cur.execute(sql)
        except Exception as e:
        #else:
            self.logger.error(e)
            self.logger.error(f'sql [{sql}]')
            self.logger.error(f'args {args}')
            traceback.print_stack()

            if self.show_message:
                messagebox.showwarning('Error', f'Database Error\n {e}')
                self.err = e
                return 0

        return self.cur.lastrowid

    def commit(self):
        """ Commit transaction """
        self.cur.execute("commit")

    def select(self, select_str, where, from_tbl, args):
        """
        Execute a SELECT statement   

        # Args:   
            select_str: string for SELECT xx   
            where: Where clause   
            from_tbl: Table name   
            args: Args tuple for Select   
            Note - ORDER clause and LIMIT clause are filled in with previously set values

        # Returns: Result list.  Self.err is set to Exception text. Shows Messagebox and/or exits on error if flags set.   
        # Raises: Nothing.  DB exceptions are suppressed. 

        """
        self.err = ''

        cur = self.conn.cursor()
        sql = f"SELECT {select_str} FROM {from_tbl} WHERE {where} {self.order_string} {self.limit_string} COLLATE BINARY"
        self.logger.debug(f'select {sql} val={args}')
        try:
            cur.execute(sql, args)
            result_list = cur.fetchall()
        except Exception as e:
            if self.show_message:
                messagebox.showwarning('Error', f'Database select error\n\n'
                f'SELECT\n {select_str}\n FROM {from_tbl} WHERE\n {where}\n'
                f'{args}\n\n {e}')
            self.err = e
            self.logger.error(e)
            self.logger.error(sql)
            traceback.print_stack()

            if self.exit_on_error:
                sys.exit()
            else:
                result_list = None
        return result_list

    def table_exists(self, table_name) -> bool:
        """
            Returns whether table exists
        # Args:
            table_name:
        # Returns:
            True if table exists
        # Raises: Nothing.  DB exceptions are suppressed
        """
        # SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';
        self.err = ''
        where = "type=? AND name=?"
        from_tbl = 'sqlite_master'
        select_str = 'name'
        args = ('table', table_name)

        cur = self.conn.cursor()

        # See if  Table exists.
        sql = f"SELECT {select_str} FROM {from_tbl} WHERE {where} {self.order_string} {self.limit_string}"

        try:
            cur.execute(sql, args)
            res = cur.fetchall()
            if len(res) > 0:
                return True
            else:
                self.logger.debug(f'{table_name} table NOT FOUND')
                return False
        except Exception as e:
            self.logger.warning(f'DB ERROR {e}')
            traceback.print_stack()

            self.err = e
            return False

    def test_database(self, select_str: str, from_tbl: str, where: str, args):
        """
            Execute a test SELECT query on database
        # Args:
            select_str: SELECT parameter
            from_tbl: table name
            where: where clause
            args: argument tuple for where clause
        # Returns:
            True if error
        # Raises: Nothing.  DB exceptions are suppressed
        """
        self.err = ''
        cur = self.conn.cursor()
        sql = f"SELECT {select_str} FROM {from_tbl} WHERE {where} {self.order_string} {self.limit_string}"

        self.logger.debug(f'db test sql={sql} args=[{args}]')
        try:
            cur.execute(sql, args)
            cur.fetchall()
            self.logger.debug('DB no error')
            return False
        except Exception as e:
            self.logger.warning(f'DB ERROR {e}')
            self.err = e
            return True

    def set_speed_pragmas(self):
        """
        Set DB pragmas for speed.  **Use with caution as these can lead to corruption!**   
        'PRAGMA temp_store = memory'   
        'PRAGMA journal_mode = off'   
        'PRAGMA locking_mode = exclusive'   
        'PRAGMA synchronous = 0   
        """
        self.logger.debug('Database pragmas set for speed')
        self.conn.isolation_level = None
        for txt in ['PRAGMA temp_store = memory',
                    'PRAGMA journal_mode = off',
                    'PRAGMA locking_mode = exclusive',
                    'PRAGMA synchronous = 0',
                    'PRAGMA case_sensitive_like = ON'
                    ]:
            self.set_pragma(txt)

    def set_optimize_pragma(self):
        """
        Set 'PRAGMA optimize'
        """
        self.logger.debug(' Database Optimize pragma')
        for txt in ['PRAGMA optimize',
                    ]:
            self.set_pragma(txt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""    Holds the details about a Location: Name,state/province, etc as well as
    the lookup results """
import argparse
import logging
import re
from typing import List, Tuple

from geodata import GeoUtil, Normalize, ArgumentParserNoExit, GeoSearch, GeoDB


# What type of entity is this place?
class PlaceType:
    COUNTRY = 0
    ADMIN1 = 1
    ADMIN2 = 2
    CITY = 3
    PREFIX = 4
    ADVANCED_SEARCH = 5


place_type_name_dict = {
    PlaceType.COUNTRY        : 'Country',
    PlaceType.ADMIN1         : 'STATE/PROVINCE',
    PlaceType.ADMIN2         : 'COUNTY',
    PlaceType.CITY           : ' ',
    PlaceType.ADVANCED_SEARCH: ' '
    }


class Loc:
    """
    Holds the details about a Location: Name, county, state/province, country, lat/long as well as
    the lookup results
    Parses a name into Loc items (county, state, etc)
    """

    def __init__(self):
        """
        Init
        """
        self.logger = logging.getLogger(__name__)
        self.original_entry: str = ""
        self.lat: float = float('NaN')  # Latitude
        self.lon: float = float('NaN')  # Longitude
        self.country_iso: str = ""  # Country ISO code
        self.country_name: str = ''
        self.city: str = ""  # City or entity name
        self.admin1_name: str = ""  # Admin1 (State/province/etc)
        self.admin1_id: str = ""  # Admin1 Geoname ID
        self.admin2_name: str = ""  # Admin2 (county)
        self.admin2_id = ""  # Admin2 Geoname ID
        self.prefix: str = ""  # Prefix (entries prepended before geoname location)
        self.feature: str = ''  # Geoname feature code
        self.place_type: int = PlaceType.COUNTRY  # Is this a Country , Admin1 ,admin2 or city?
        self.geoid: str = ''  # Geoname GEOID
        self.enclosed_by = ''   # The entity that encloses this.  E.g United States encloses Texas
        self.updated_entry = ''
        self.score = 100.0

        # Lookup result info
        self.status: str = ""
        self.status_detail: str = ""
        self.result_type: int = GeoUtil.Result.NO_MATCH  # Result type of lookup
        self.result_type_text: str = ''  # Text version of result type
        self.georow_list: List = []  # List of items that matched this location
        self.event_year: int = 0
        self.geo_db = None

    def clear(self):
        # Place geo info
        self.original_entry: str = ""
        self.lat: float = float('NaN')  # Latitude
        self.lon: float = float('NaN')  # Longitude
        self.country_iso: str = ""  # Country ISO code
        self.country_name: str = ''
        self.city: str = ""  # City or entity name
        self.admin1_name: str = ""  # Admin1 (State/province/etc)
        self.admin2_name: str = ""  # Admin2 (county)
        self.admin1_id: str = ""  # Admin1 Geoname ID
        self.admin2_id = ""  # Admin2 Geoname ID
        self.prefix: str = ""  # Prefix (entries before city)
        self.feature: str = ''  # Geoname feature code
        self.place_type: int = PlaceType.COUNTRY  # Is this a Country , Admin1 ,admin2 or city?
        self.geoid: str = ''  # Geoname GEOID
        self.enclosed_by = ''
        self.updated_entry = ''
        self.score = 100.0

        # Lookup result info
        self.status: str = ""
        self.status_detail: str = ""
        self.result_type: int = GeoUtil.Result.NO_MATCH  # Result type of lookup
        self.result_type_text: str = ''  # Text version of result type
        self.georow_list: List[Tuple] = [()]  # List of items that matched this location

        self.georow_list.clear()

    def parse_place(self, place_name: str, geo_db:GeoDB.GeoDB):
        """
            Given a comma separated place name,   
            parse into its city, admin1, country and type of entity (city, country etc)   
        #Args:   
            place_name: The place name to parse   
            geo_files: GeodataBuild instance   
        #Returns:   
            Fields in Loc (city, adm1, adm2, iso) are updated based on parsing. self.status has Result status code   
        """
        self.geo_db = geo_db
        self.logger.debug(f'PARSE {place_name}\n')
        self.clear()
        self.original_entry = place_name

        # Convert open-brace and open-paren to comma.  close brace/paren will be stripped by normalize()
        name = re.sub(r'\[', ',', place_name)
        name = re.sub(r'\(', ',', name)

        tokens = name.split(",")
        if len(tokens[-1]) == 0:
            # Last item is blank, so remove it
            tokens = tokens[:-1]

        token_count = len(tokens)
        self.place_type = PlaceType.CITY

        # First, try to parse and validate State/Province, and Country from last two tokens  
        # If one other token, parse as city
        # If two other tokens, parse as city, admin2
        # First two tokens are also copied to prefix.
        # Place type is the leftmost item we found - either City, Admin2, Admin2, or Country
        # If '--' in name, then extract advanced search options 

        if '--' in place_name:
            # Advanced Search - Pull out filter flags if present
            self.logger.debug('filter')
            self.get_filter_parameters(place_name)
            return
        
        if token_count > 0:
            #  COUNTRY - right-most token should be country
            self.country_name = Normalize.normalize(tokens[-1], False)

            # Validate country
            #self.logger.debug(f'1) Find COUNTRY [{self.country_name}] *******')
            self.country_iso = geo_db.s.get_country_iso(self.georow_list, self.country_name, self)  # Get Country country_iso
            if self.country_iso != '':
                self.place_type = PlaceType.COUNTRY
            else:
                # Last token is not COUNTRY.
                # Append dummy token  so we now have <tokens>, x
                tokens.append('_')
                token_count = len(tokens)
                self.result_type = GeoUtil.Result.NO_COUNTRY
                self.country_name = ''
            #self.logger.debug(f'ISO =[{self.country_iso}]')
        if token_count > 1:
            #  See if 2nd to last token is Admin1
            val = tokens[-2]
            #self.logger.debug(f'ADM1 tkn-2 [{val}]')
            self.admin1_name = Normalize.admin1_normalize(val, self.country_iso)

            if len(self.admin1_name) > 0:
                # Lookup Admin1
                #self.logger.debug(f'2) Find ADMIN1 [{self.admin1_name}] *******')
                row_list = []
                geo_db.s.get_admin1_id(self.admin1_name, self, row_list)
                if self.admin1_id != '':
                    # Found Admin1
                    self.place_type = PlaceType.ADMIN1
                    self.georow_list = row_list
                    self.admin1_name = geo_db.s.get_admin1_name(self.admin1_id, self.country_iso)
                    #self.logger.debug(f'adm1 nm={self.admin1_name}')
                    self.result_type = GeoUtil.Result.PARTIAL_MATCH
                    # Get country if blank
                    row_list = []
                    if self.country_name == '':
                        self.country_name = geo_db.s.get_country_name(self.country_iso, row_list) 
                else:
                    # Last token is not Admin1 - append dummy token so we have <tokens>, admin1, country
                    self.admin1_name = ''
                    # Add dummy token for admin1 position
                    tokens.insert(-1, '_')
                    # token_count = len(tokens)
            else:
                tokens[-2] = '_'

        # Last two tokens are now Admin1, Country (although they may be '_')
        # If >2 tokens:  Put first non-blank token in City and in Prefix
        # If >3 tokens:  Put second non-blank token in Admin2 and also append to Prefix

        # Remove all blank tokens
        tokens = [x for x in tokens if x]
        token_count = len(tokens)

        if token_count >= 3:
            #  Possible Formats: City, Admin1, Country or  Admin2, Admin1, Country
            #  Take first tkn as city
            self.city = Normalize.normalize(tokens[0], False)
            self.place_type = PlaceType.CITY

            # Also place token[0] into Prefix
            if '*' not in tokens[0]:
                self.prefix = str(tokens[0].strip(' '))
                
        if token_count >= 4:
            #  Admin2 is 2nd.  Note -  if Admin2 isnt found, it will look it up as city
            
            if GeoUtil.is_street(tokens[-4].lower()):
                #  Format: Prefix, City, Admin1, Country
                self.city = Normalize.normalize(tokens[-3], False)
            else:
                #  Format: City, Admin2, Admin1, Country
                self.admin2_name = Normalize.normalize(tokens[-3], False)
                self.city = Normalize.normalize(tokens[-4], False)
                
            self.place_type = PlaceType.CITY

            # put token[0] and  token[1] into Prefix
            if '*' not in tokens[1]:
                self.prefix = str(tokens[0].strip(' ')) + ' ' + str(tokens[1].strip(' '))

        self.prefix = Normalize.normalize(self.prefix, False)
        row_list = []
        # fill in country name if still missing - finding Admin1 will find country ISO
        if self.country_name == '' and self.country_iso != '':
            self.country_name = geo_db.s.get_country_name(self.country_iso, row_list)

        self.logger.debug(f"    ======= PARSED: {place_name} \nCity [{self.city}] Adm2 [{self.admin2_name}]"
                          f" Adm1 [{self.admin1_name}] adm1_id [{self.admin1_id}] Cntry [{self.country_name}] Pref=[{self.prefix}]"
                          f" type_id={self.place_type}\n")
        return

    def get_filter_parameters(self, place_name):
        """
            Parse search parameters from place_name and place in appropriate fields.
            Place_name format is:  town, --feature=XXX,--iso=XXX
        # Args:   
            place_name: town, --feature=XXX,--iso=XXX   
        # Returns:   
            Loc fields updated: country_iso, and feature are updated   
        """
        # Separate out arguments
        tokens = place_name.split(",")
        args = []
        for tkn in tokens:
            if '--' in tkn:
                args.append(tkn.strip(' '))

        # Parse options in place name
        parser = ArgumentParserNoExit.ArgumentParserNoExit(description="Parses command.")
        parser.add_argument("-f", "--feature", help=argparse.SUPPRESS)
        parser.add_argument("-i", "--iso", help=argparse.SUPPRESS)
        parser.add_argument("-c", "--country", help=argparse.SUPPRESS)
        self.logger.debug(f'Args {args}' )

        try:
            options = parser.parse_args(args)
            self.city = Normalize.normalize(tokens[0], False)
            self.place_type = PlaceType.ADVANCED_SEARCH
            self.logger.debug(options)

            if options.iso:
                self.country_iso = options.iso.lower()
                self.logger.debug(f'iso {self.country_iso}')
            if options.country:
                self.country_iso = options.country.lower()
            if options.feature:
                self.feature = options.feature.upper()
                self.logger.debug(f'ft {self.feature}')
        except Exception as e:
            self.logger.debug(e)
            
        self.logger.debug(f"    ======= PARSE ADV: {place_name} \nCity [{self.city}] Adm2 [{self.admin2_name}]"
                          f" Adm1 [{self.admin1_name}] adm1_id [{self.admin1_id}] Cntry [{self.country_name}] Pref=[{self.prefix}]"
                          f" type_id={self.place_type} iso={self.country_iso} feat={self.feature}")

    def get_status(self) -> str:
        self.logger.debug(f'status=[{self.status}]')
        return self.status

    def set_types_as_string(self):
        # Ensure all items are string type
        self.city = str(self.city)
        self.admin1_name = str(self.admin1_name)
        self.admin2_name = str(self.admin2_name)
        self.prefix = str(self.prefix)

    def add_commas(self, txt) -> str:
        if txt == '':
                return ''
        else:
            return f'{txt}, '

    @staticmethod
    def lowercase_match_group(matchobj):
        return matchobj.group().lower()

    @staticmethod
    def capwords(nm):
        # Change from lowercase to Title Case but fix the title() apostrophe bug
        if nm is not None:
            # Use title(), then fix the title() apostrophe defect
            nm = nm.title()

            # Fix handling for contractions not handled correctly by title()
            poss_regex = r"(?<=[a-z])[\']([A-Z])"
            nm = re.sub(poss_regex, Loc.lowercase_match_group, nm)

        return nm

    def get_long_name(self, replace_dct) -> str:
        """
        Take the fields in a Place and build full name.  e.g.  city,adm2,adm1,country name   
        Prefix is NOT included.  Text also has replacements from dictionary applied   
        #Args:   
            replace_dct: Dictionary of text replacements.  'Regex':'replacement'   
        #Returns:  
            long name  
        """
        city = self.add_commas(self.city)
        admin2 = self.add_commas(self.admin2_name)
        admin1 = self.add_commas(self.admin1_name)

        if self.place_type == PlaceType.COUNTRY:
            nm = f"{self.country_name}"
        elif self.place_type == PlaceType.ADMIN1:
            nm = f"{admin1}{self.country_name}"
        elif self.place_type == PlaceType.ADMIN2:
            nm = f"{admin2}{admin1}{self.country_name}"
        else:
            nm = f"{city}{admin2}{admin1}{str(self.country_name)}"

        # normalize prefix
        self.prefix = Normalize.normalize(self.prefix, False)

        if len(self.prefix) > 0:
            self.prefix_commas = ', '
        else:
            self.prefix_commas = ''

        nm = Loc.capwords(nm)

        # Perform any text replacements user entered into Output Tab
        if replace_dct:
            for key in replace_dct:
                nm = re.sub(key, replace_dct[key], nm)

        return nm
    
    def get_display_name(self, replace_dct) -> str:
        """
        Take the fields in a Place and build full name.  See if alternate name is available
        e.g.  city,adm2,adm1,country name   
        Prefix is NOT included.  Text also has replacements from dictionary applied   
        #Args:   
            replace_dct: Dictionary of text replacements.  'Regex':'replacement'   
        #Returns:  
            long name  
        """
        
        # See if there is an alternate name for this
        #city, lang = self.geo_db.s.get_alternate_name(self.geoid)
        #if city == '':
        city = self.city
        city = self.add_commas(city)
        admin2 = self.add_commas(self.admin2_name)
        admin1 = self.add_commas(self.admin1_name)

        if self.place_type == PlaceType.COUNTRY:
            nm = f"{self.country_name}"
        elif self.place_type == PlaceType.ADMIN1:
            nm = f"{admin1}{self.country_name}"
        elif self.place_type == PlaceType.ADMIN2:
            nm = f"{admin2}{admin1}{self.country_name}"
        else:
            nm = f"{city}{admin2}{admin1}{str(self.country_name)}"

        # normalize prefix
        self.prefix = Normalize.normalize(self.prefix, False)

        if len(self.prefix) > 0:
            self.prefix_commas = ', '
        else:
            self.prefix_commas = ''

        nm = Loc.capwords(nm)

        # Perform any text replacements user entered into Output Tab
        if replace_dct:
            for key in replace_dct:
                nm = re.sub(key, replace_dct[key], nm)

        return nm

    def get_five_part_title(self):
        # Returns a five part title string and tokenized version:
        #     prefix,city,county,state,country  (prefix plus long name)

        # Normalize country name
        country, modified = Normalize.country_normalize(self.country_name)
        full_title = self.prefix + ' ,' + f"{self.city}, {self.admin2_name}, {self.admin1_name}, {str(country)}"
        return full_title

    def set_type_from_feature(self):
        # Set place type based on DB response feature code
        if self.feature == 'ADM0':
            self.place_type = PlaceType.COUNTRY
        elif self.feature == 'ADM1':
            self.place_type = PlaceType.ADMIN1
        elif self.feature == 'ADM2':
            self.place_type = PlaceType.ADMIN2
        else:
            self.place_type = PlaceType.CITY
        if len(self.prefix) > 0:
            self.place_type = PlaceType.PREFIX
        return self.place_type

    def set_place_type(self):
        # Set place type based on parsing results
        self.place_type = PlaceType.CITY
        if len(str(self.country_name)) > 0:
            self.place_type = PlaceType.COUNTRY
        if len(self.admin1_name) > 0:
            self.place_type = PlaceType.ADMIN1
        if len(self.admin2_name) > 0:
            self.place_type = PlaceType.ADMIN2
        if len(self.city) > 0:
            self.place_type = PlaceType.CITY

    def set_place_type_text(self):
        # Set result_type_text based on place type
        if self.result_type == GeoUtil.Result.NO_COUNTRY:
            self.result_type_text = 'Country'
        elif self.place_type == PlaceType.COUNTRY:
            self.result_type_text = 'Country'
        elif self.place_type == PlaceType.ADMIN1:
            self.result_type_text = self.get_district1_type(self.country_iso)
        elif self.place_type == PlaceType.ADMIN2:
            self.result_type_text = 'County'
        elif self.place_type == PlaceType.CITY:
            self.result_type_text = self.get_type_name(self.feature)
        elif self.place_type == PlaceType.PREFIX:
            self.result_type_text = 'Place'

    def remove_old_fields(self):
        # Remove fields that are unused by this place type
        if self.place_type == PlaceType.COUNTRY:
            # self.prefix = ''
            self.city = ''
            self.admin2_name = ''
            self.admin1_name = ''
        elif self.place_type == PlaceType.ADMIN1:
            # self.prefix = ''
            self.city = ''
            self.admin2_name = ''
        elif self.place_type == PlaceType.ADMIN2:
            # self.prefix = ''
            self.city = ''
        elif self.place_type == PlaceType.CITY:
            # self.prefix = ''
            pass

    @staticmethod
    def sort_words(words: str) -> str:
        word_list: list = words.split(' ')
        word_list.sort()
        return ' '.join(word_list)

    @staticmethod
    def get_soundex_by_word(text: str) -> str:
        result_sdx = ''
        text_words = text.split(' ')
        for word in text_words:
            if len(result_sdx) == 0:
                result_sdx = GeoSearch.get_soundex(Loc.sort_words(word))
            else:
                result_sdx += ' ' + GeoSearch.get_soundex(Loc.sort_words(word))
        return result_sdx

    @staticmethod
    def matchscore_prefix(pref: str, result: str) -> str:
        """
        Remove any items from prefix that are in match result.  Remove *   
        #Args:   
            pref:   
            result:   

        #Returns:  Prefix with words removed   
        """
        new_prfx = pref.lower()
        new_prfx = new_prfx.strip(' ')
        prefix_parts = new_prfx.split(',')

        result = result.lower()
        result_parts = result.split(',')

        # Walk thru each segment in result
        for result_segment_idx, result_segment in enumerate(result_parts):
            result_sdx = ' ' + Loc.get_soundex_by_word(result_segment) + ' '
            result_segment = ' ' + result_segment + ' '
            # Walk thru each segment in prefix
            for prefix_segment_idx, prefix_segment in enumerate(prefix_parts):
                prefix_words = prefix_segment.split(' ')
                # Walk thru each word in prefix segment
                for pref_word_idx, prefix_word in enumerate(prefix_words):
                    prefix_sdx = ' ' + GeoSearch.get_soundex(prefix_word) + ' '
                    if len(prefix_word) < 3:
                        prefix_word += ' '
                    # Remove words in prefix that are in result_segment
                    if (prefix_word in result_segment and prefix_word != '') or (prefix_sdx in
                                                                                 result_sdx and prefix_sdx != ''):
                        result_segment = remove_item(prefix_word.strip(' '), result_segment)
                        new_prfx = remove_item(prefix_word.strip(' '), new_prfx)
                        pass
                    # See if any words in result_segment are in prefix
                    result_words = result_segment.split(' ')
                    if len(prefix_word) > 0:
                        # Walk through each word in result
                        for result_word_idx, result_word in enumerate(result_words):
                            if len(result_word) < 3:
                                result_word = result_word + ' '
                            if result_word in prefix_word and float(len(result_word)) / float(len(prefix_word)) > 0.6:
                                # Remove result_word from result and prefix
                                result_segment = remove_item(result_word, result_segment)
                                new_prfx = remove_item(result_word, new_prfx)

        res = re.sub('[,]', '', new_prfx)
        res = res.strip(' ')
        return res.strip(',')

    @staticmethod
    def prefix_cleanup(pref: str, result: str) -> str:
        """
        Cleanup prefix.  Remove any items from prefix that are in match result.  Remove *   
        #Args:   
            pref:   
            result:   

        #Returns:  Prefix with words removed

        """
        return Loc.matchscore_prefix(pref, result)

    @staticmethod
    def get_district1_type(iso) -> str:
        # Return the local country term for Admin1 district
        if iso in ["al", "no"]:
            return "County"
        elif iso in ["us", "at", "bm", "br", "de"]:
            return "State"
        elif iso in ["ac", "an", 'ao', 'bb', 'bd']:
            return "Parish"
        elif iso in ["ae"]:
            return "Emirate"
        elif iso in ["bc", "bf", "bh", "bl", "bn"]:
            return "District"
        elif iso in ["gb"]:
            return "Country"
        else:
            return "Province"

    @staticmethod
    def get_type_name(feature):
        nm = type_name.get(feature)
        if nm is None:
            nm = ''
        return nm

    def update_names(self,  dct):
        prfx = self.prefix_cleanup(self.prefix, self.get_long_name(dct))
        self.updated_entry = GeoUtil.capwords(prfx) + self.get_long_name(dct)


def remove_item(pattern, text) -> str:
    if len(pattern) < 1:
        return text
    # Find pattern in word in text and remove entire word
    segment_list = text.split(',')
    # Walk thru segments
    for seg_idx, segment in enumerate(segment_list):
        # Walk thru words
        word_list = segment.split(' ')
        for idx, word in enumerate(word_list):
            if pattern in word:
                # Remove entire word, not just pattern
                word_list[idx] = ''
        segment_list[seg_idx] = ' '.join(word_list)
    text = ','.join(segment_list)
    return text


type_name = {
    "ADM0": 'Country', "ADM1": 'City', "ADM2": 'City', "ADM3": 'City', "ADM4": 'City', "ADMF": 'City',
    "CH"  : 'Church', "CSTL": 'Castle', "CMTY": 'Cemetery', "EST": 'Estate', "HSP": 'Hospital',
    "HSTS": 'Historic', "ISL": 'Island', "MSQE": 'Mosque', "MSTY": 'Monastery', "MT": 'Mountain', "MUS": 'Museum', "PAL": 'Palace',
    "PPL" : 'City', "PPLA": 'City', "PPLA2": 'City', "PPLA3": 'City', "PPLA4": 'City',
    "PPLC": 'City', "PPLG": 'City', "PPLH": 'City', "PPLL": 'Village', "PPLQ": 'City', "PPLX": 'City',
    "PRK" : 'Park', "PRN": 'Prison', "PRSH": 'Parish', "RUIN": 'Ruin',
    "RLG" : 'Religious', "STG": '', "SQR": 'Square', "SYG": 'Synagogue', "VAL": 'Valley', "PP1M": 'City', "PP1K": 'City'
    }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""    Example program for Geodata gazeteer.  
"""
import logging
import os
import sys
from pathlib import Path

import numpy as np
from scipy.optimize import minimize

from geodata import Geodata, Loc


class Optimize:
    """
    Example program for Geodata gazeteer.  
    0. pip3 install geodata
    1. Create folder in home directory:  example/cache  
    2. Download gb.txt from geonames.org and place in example/cache.  (This just has Great Britain data)  
    3. Run example.py  
    """

    def __init__(self):
        # Set up standard logging.  
        logging.getLogger(__name__)
        fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.INFO, stream=sys.stdout, format=fmt)  # Change this to logging.DEBUG for more detail
        self.result_cache = {}

        # Initialize
        directory = os.path.join(str(Path.home()), "Documents", "geoname_data")
        self.geodata = Geodata.Geodata(directory_name=directory, display_progress=None,
                                       show_message=True, exit_on_error=True,
                                       languages_list_dct={'en'},
                                       feature_code_list_dct=features,
                                       supported_countries_dct={'fr', 'gb', 'ca'})

        # Open Geoname database - city names, lat/long, etc.  Create database if not found
        error = self.geodata.open(repair_database=True, query_limit=105)
        if error:
            print(f"Missing geoname Files in {directory}: download gb.txt or allcountries.txt from geonames.org")
            raise ValueError('Missing files from geonames.org')

    def find_match(self, location_name):
        # Create Location instance.  This will hold search parameters and result
        place: Loc.Loc = Loc.Loc()

        # Find best match
        # Check cache
        if self.result_cache.get(location_name):
            return self.result_cache[location_name]
        else:
            # Not in cache, lookup in DB
            self.geodata.find_best_match(location=location_name, place=place)
            self.result_cache[location_name] = place
            return place

    def get_place_lookup_score(self, location_name):
        target_place = Loc.Loc()
        # Find best match
        place = self.find_match(location_name)
        target_place.parse_place(place_name=location_name, geo_db=self.geodata.geo_build)

        if len(place.georow_list) > 0:
            # Create full name for result
            name = f'{place.get_long_name(None)}'
            print(f'\n   Best match for {location_name}:\n {name}  Prefix=[{place.prefix}{place.prefix_commas}] Score= {place.score:.1f}')
            return self.geodata.geo_build.geodb.match.match_score(target_place, place)
            # return place.score
        else:
            if place.result_type == Geodata.GeoUtil.Result.NOT_SUPPORTED:
                print(f'   NO match for {location_name}:\n Country {place.country_name} not in supported country list')
            else:
                print(f'   NO match for {location_name}:\n')
            return 100.0

    def calculate(self, arg):
        # Initialize
        results = []

        args = list(arg)

        print(f'==== TRYING <<{args[0:3]}>> pref={args[3]}')

        # Weighting for each input term match - prefix, city, adm2, adm1, country
        self.geodata.geo_build.geodb.match.set_weighting(token_weight=args[0:3], prefix_weight=args[3], feature_weight=0.05)

        for idx, name in enumerate(locations):
            delta = abs(ex.get_place_lookup_score(locations[idx]) - scores[idx])
            results.append(delta)

        score = rmse(y=results, y_pred=scores)
        print(f'OVERALL={score:.1f}\nCty,State,Ctry,Prefix, \n')
        return score


def rmse(y, y_pred):
    # Root mean square error
    sqr = np.square(np.array(y) - np.array(y_pred))
    return np.sqrt(np.mean(sqr))


# Try a few different locations
locations: [str] = [
    #'edinburgh ,edinburgh,scotland,united kingdom',  #1
    'edinburgh ,edinburgh,qqqqqqqq,united kingdom',  #2
    'edinburgh ,qqqqqqqqq,scotland,united kingdom',  #3
    #'qqqqqqqqq,edinburgh ,edinburgh,scotland,united kingdom',  #4
    #'aaaa, edinburgh ,edinburgh,scotland,united kingdom',  # 5
    # '12 qqqqqqqqqqq, edinburgh ,edinburgh,scotland',  # 6
    # '12 baker st,edinburgh,ile de france,scotland,united kingdom',  # 7
    # 'eddinburg ,,scotland',  # misspelled #8
    #'Londorn qwx, zxq, q, United Kingdom', #9
    #'Londorn qwx, zxq, q, United Kingdom', #10
    #'Londorn qwx, zxq, q, United Kingdom', #11
    ]

scores = [
    #-12.0, #1
    -9.0, #2
    -4, #3
    #15.0, #4
    #10.0, #5
    # 25.0, #6
    # 13.0, #7
    # 13.0, #8
    #85.0, #9
    #85.0, #10
    #85.0, #11
    ]

# Geoname feature types to add to database.  Other feature types will be ignored.
features = {"ADM1", "ADM2", "ADM3", "ADM4", "ADMF", "CH", "CSTL", "CMTY", "EST ", "HSP", "FT",
            "HSTS", "ISL", "MSQE", "MSTY", "MT", "MUS", "PAL", "PPL", "PPLA", "PPLA2", "PPLA3", "PPLA4",
            "PPLC", "PPLG", "PPLH", "PPLL", "PPLQ", "PPLX", "PRK", "PRN", "PRSH", "RUIN", "RLG", "STG", "SQR", "SYG", "VAL"}

if __name__ == "__main__":
    # Initialize
    ex = Optimize()

    #       Conty,      State,      Ctry,       Prefix,      (Input=1 - feat - result)
    bnds = ((0.02, 1.0), (0.02, 1.0), (0.02, 1.0), (0, 10.0),)
    # county, state, country, prefix
    val = np.array([0.05, 0.05, 1., 2.0])
    res = minimize(ex.calculate, val, method='TNC', bounds=bnds)

    print(res.x)
